#!/usr/bin/python3

#
# QA Cafe NTA1000 Create Image Utility
# Copyright 2021, QA Cafe
#
# This script will look on the QA Cafe server and download the latest
# USB image used for reimaging the NTA1000 internal drives, burn the
# image to a USB drive connected to the system, provide instructions on
# how to boot to the USB drive, poweroff.  The script will check if the
# latest USB image has already been downloaded and has the correct size
# and MD5 checksum, skipping the download if so.

import argparse
import sys
import subprocess
import re
import os
import time
import urllib.request
import shutil
import queue
import threading
import math
import tempfile

def main():

    # Handle command line.
    parser = argparse.ArgumentParser(description='NTA1000 Create Image Utility.')
    parser.add_argument('--offline', action='store_true', help='create the bootable USB drive on another computer, useful if your NTA1000 is not connected to the internet')
    args = parser.parse_args()

    verify_user()
    splash()

    # Determine NTA1000 system.
    nta1000_system = get_nta1000_system(args.offline)
    if nta1000_system is None:
        print('The program was unable to identify the NTA1000 hardware.')
        print('If you are running this program on a different system, you need to specify the --offline option.')
        sys.exit(1)

    # Determine the USB drive.
    usb_drive = get_usb_drive()
    if usb_drive is None:
        sys.exit(1)

    # Unmount any mounted partitions on the USB drive.
    if not unmount_usb_drive(usb_drive):
        sys.exit(1)

    # Get/verify the appropriate image file.
    (local_image_file, local_image_size) = get_local_image_file(nta1000_system)
    if local_image_file is None:
        sys.exit(1)

    # Burn the image file to the USB drive.
    if not burn_image(local_image_file, local_image_size, usb_drive):
        sys.exit(1)

    # Tell the user how to boot from the USB drive.
    display_usb_boot_instructions(nta1000_system, args.offline)

    # Power off, unless this is an offline run.
    if not args.offline:
        input('Hit Enter to power off.')
        print()
        print()
        print()
        print('Powering off...')
        print()
        time.sleep(5)
        os.system('poweroff')

################################################################################

def verify_user():

    if os.geteuid() != 0:
        print('This script must be run as root.')
        sys.exit(1)

################################################################################

def splash():

    subprocess.run(['clear'])
    print('QA Cafe NTA1000 Disk Drive Reimage Utility')
    print('------------------------------------------')
    print()
    print('This utility will guide you through the process of reimaging your internal')
    print('drives.  You will be given the choice to rebuild the OS drive or the')
    print('data drive (or both).  CDRouter data (Devices, Configs, Packages, Results)')
    print('are stored on the data drive.')
    print()
    print('Reimaging a drive will delete any existing data on that drive.  Be sure you')
    print('have backed up anything you want to restore later.')
    print()
    print('This process requires a 4GB (or larger) USB drive.  Be aware that any data')
    print('currently on the USB drive will also be deleted.')
    print()
    print('If your NTA1000 is connected to the internet, you should be running this')
    print('program on the NTA1000.  Make sure the USB drive is plugged into one of the')
    print('USB ports on the front of the system before continuing.')
    print()
    print('If your NTA1000 is not connected to the internet, you should be running this')
    print('program on a different CentOS or Rocky Linux system that has an internet')
    print('connection.  Be sure to specify the "--offline" option.  Make sure the USB')
    print('drive is plugged into one of the USB ports on your system before continuing.')
    print('Once the USB drive has been created, you will be prompted to remove it from')
    print('your system, plug it into your NTA1000, and boot to it.')
    print()
    print('Please direct any questions to support@qacafe.com.')
    print()

    choice = None
    while(True):
        choice = input('Continue? (y,n) ')
        if choice in ['y', 'n']:
            break
    print()
    if choice == 'n':
        sys.exit(0)

################################################################################

def get_nta1000_system(offline):

    nta1000_system = None

    if offline:

        # Determine the system from the NTA serial number.
        print('The NTA1000 serial number is needed in order to determine the correct')
        print('image to use.  NTA1000 serial numbers always start with "NTA1000-".')
        print('This can be found on the rear of the unit.  On some systems, it\'s also')
        print('displayed on the front LCD panel')
        print()
        while True:
            nta1000_sn = input('Please enter the serial number of the NTA1000 system -> ')
            if re.match('NTA1000-10\d\d\d', nta1000_sn):
                if nta1000_sn[10:11] in ['0', '1']:
                    nta1000_system = 'FWA8206'
                    break
                elif nta1000_sn[10:11] in ['2', '3', '4']:
                    nta1000_system = 'FWA8207'
                    break
                elif nta1000_sn[10:11] == '5':
                    nta1000_system = 'FWA8308'
                    break
            elif re.match('NTA1000v5-\d\d\d', nta1000_sn):
                nta1000_system = 'FWA8308'
                break
            elif re.match('NTA1000v[67]G?-\d\d\d', nta1000_sn):
                nta1000_system = 'FWA8408'
                break
            print('Invalid NTA1000 serial number.')

        print()

    else:

        # Get known ethernet interfaces present on the PCI bus.
        ethernet_I217_LM = 0
        ethernet_I210    = 0
        ethernet_82578DM = 0
        ethernet_82574L  = 0

        try:
            output = subprocess.run(['lspci'], check=True, universal_newlines=True, stdout=subprocess.PIPE).stdout
            lines = output.splitlines()
            for line in lines:
                if re.match('[0-9a-fA-F:\.]+ Ethernet controller: Intel Corporation Ethernet Connection I217-LM', line):
                    ethernet_I217_LM += 1
                elif re.match('[0-9a-fA-F:\.]+ Ethernet controller: Intel Corporation I210 Gigabit Network Connection', line):
                    ethernet_I210 += 1
                elif re.match('[0-9a-fA-F:\.]+ Ethernet controller: Intel Corporation 82578DM Gigabit Network Connection', line):
                    ethernet_82578DM += 1
                elif re.match('[0-9a-fA-F:\.]+ Ethernet controller: Intel Corporation 82574L Gigabit Network Connection', line):
                    ethernet_82574L += 1

            # FWA8408
            if (ethernet_I217_LM == 0 and
                ethernet_I210    == 8 and
                ethernet_82578DM == 0 and
                ethernet_82574L  == 0):

                nta1000_system = 'FWA8408'

            # FWA8308
            elif (ethernet_I217_LM == 1 and
                ethernet_I210    == 7 and
                ethernet_82578DM == 0 and
                ethernet_82574L  == 0):

                nta1000_system = 'FWA8308'

            # FWA8207
            elif (ethernet_I217_LM == 0 and
                ethernet_I210    == 0 and
                ethernet_82578DM == 1 and
                ethernet_82574L  == 6):

                nta1000_system = 'FWA8207'

            # FWA8206
            elif (ethernet_I217_LM == 0 and
                ethernet_I210    == 0 and
                ethernet_82578DM == 0 and
                ethernet_82574L  == 6):

                nta1000_system = 'FWA8206'

        except Exception as e:
            print('Error: {}'.format(e))

    return nta1000_system

################################################################################

def get_usb_drive():

    usb_drive = None
    usb_drive_size = None

    try:
        output = subprocess.run(['lsblk', '-ldnbo', 'TYPE,RM,ROTA,STATE,TRAN,NAME,SIZE'], check=True, universal_newlines=True, stdout=subprocess.PIPE).stdout
        lines = output.splitlines()
        for line in lines:
            words = line.split()
            if len(words) == 7 and words[0] == 'disk' and words[1] == '1' and words[2] == '1' and words[3] == 'running' and words[4] == 'usb':
                usb_drive = '/dev/{}'.format(words[5])
                usb_drive_size = int(words[6])
                break

        if usb_drive is None:
            print('USB drive not found.')
        else:
            print('USB drive found at {}'.format(usb_drive))
            if usb_drive_size < 4 * 1000 * 1000 * 1000:
                print('The USB drive ({}) is too small.'.format(usb_drive))
                print('Please use a USB drive at least 4GB in size.')
                usb_drive = None

    except Exception as e:
        print('Error: {}'.format(e))

    return usb_drive

################################################################################

def unmount_usb_drive(usb_drive):

    retval = True

    # Find any and all mounted partitions on this drive and unmount them.
    try:
        output = subprocess.run(['lsblk', '-lno', 'MOUNTPOINT', usb_drive], check=True, universal_newlines=True, stdout=subprocess.PIPE).stdout
        lines = output.splitlines()
        for line in lines:
            words = line.split()
            if len(words) == 1:
                subprocess.run(['umount', words[0]], check=True)

    except Exception as e:
        print('Error: {}'.format(e))
        retval = False

    return retval

################################################################################

def get_local_image_file(nta1000_system):

    local_image_file   = None
    local_image_size   = None
    local_image_md5sum = None

    q = queue.Queue(maxsize=1)

    try:
        image_file   = None
        image_size   = None
        image_md5sum = None

        # Look on the server to see what is available.
        print('Checking the QA Cafe server for the latest image...')
        local_image_info_file = os.path.join('/tmp', 'nta1000-latest-image.txt')
        with open(local_image_info_file, 'wb') as f:
            with urllib.request.urlopen('http://repo.qacafe.com/systemimages/nta1000-latest-image.txt') as r:
                shutil.copyfileobj(r, f)
        with open(local_image_info_file, 'r') as f:
            for line in f:
                words = line.split()
                if len(words) == 4:
                    if words[0] == nta1000_system:
                        image_file   = words[1]
                        image_size   = int(words[2])
                        image_md5sum = words[3]
                        break
        if image_file is None or image_size is None or image_md5sum is None:
            raise Exception('Could not get image information from QA Cafe server.')

        print('    Image file = {} ({:,} bytes)'.format(image_file, image_size))
        local_image_file = os.path.join('/tmp', image_file)
        local_image_size = 0

        # Has it already been downloaded and everything looks good?  If not, download it now.
        download_needed = True
        print('Checking if this image file has already been downloaded...', end=' ', flush=True)
        if not os.path.isfile(local_image_file):
            print('no')
        else:
            print('yes')
            print('    Verifying file size...', end=' ', flush=True)
            local_image_size = os.path.getsize(local_image_file)
            if local_image_size != image_size:
                print('incorrect file size')
            else:
                print('OK')
                print('    Verifying file checksum...', end=' ', flush=True)
                if get_md5sum(local_image_file) != image_md5sum:
                    print('bad checksum')
                else:
                    print('OK')
                    download_needed = False

        if download_needed:
            print('Downloading image file to {}'.format(local_image_file))

            # Create and start the download_progress_thread.
            download_progress_thread = threading.Thread(target=download_progress, args=(local_image_file, image_size, q))
            download_progress_thread.start()

            # Start the download.
            with open(local_image_file, 'wb') as f:
                with urllib.request.urlopen('http://repo.qacafe.com/systemimages/{}'.format(image_file)) as r:
                    shutil.copyfileobj(r, f)

            # Kill download_progress_thread and wait for it to die.
            q.put('kill')
            download_progress_thread.join()

            # Does everything look OK?
            if not os.path.isfile(local_image_file):
                print('Missing local image file: {}'.format(local_image_file))
                local_image_file = None
            else:
                local_image_size = os.path.getsize(local_image_file)
                if local_image_size != image_size:
                    print('Image file size mismatch, expecting {}, got {:,}'.format(image_size, local_image_size))
                    local_image_file = None
                else:
                    local_image_md5sum = get_md5sum(local_image_file)
                    if local_image_md5sum != image_md5sum:
                        print('Image file md5sum mismatch, expecting {}, got {}'.format(image_md5sum, local_image_md5sum))
                        local_image_file = None
                        local_image_size = None

    except Exception as e:
        print('Error: {}'.format(e))
        if download_progress_thread.is_alive():
            q.put('kill')
            download_progress_thread.join()
        local_image_file = None
        local_image_size = None

    return (local_image_file, local_image_size)

################################################################################

def get_md5sum(local_image_file):

    md5sum = None

    try:
        output = subprocess.run(['md5sum', local_image_file], check=True, universal_newlines=True, stdout=subprocess.PIPE).stdout
        words = output.split()
        if len(words) == 2:
            md5sum = words[0]

    except Exception as e:
        print('Error: {}'.format(e))

    return md5sum

################################################################################

def burn_image(local_image_file, local_image_size, usb_drive):

    retval = False
    q = queue.Queue(maxsize=1)

    try:
        # Make sure the USB drive is not mounted.
        with open('/etc/mtab', 'r') as f:
            for line in f:
                words = line.split()
                if re.match(usb_drive, words[0]):
                    print('USB drive is mounted.  Attempting to unmount...')
                    subprocess.run(['umount', words[0]], check=True)

        # Burn it!  Burn a chunk at a time so we can track progress.
        print('Burning image file to the USB drive.')
        for i in range(0, math.ceil(local_image_size / (20 * 1024 * 1024))):
            subprocess.run(['dd', 'if={}'.format(local_image_file), 'of={}'.format(usb_drive), 'bs=20M', 'conv=fsync', 'count=1', 'skip={}'.format(i), 'seek={}'.format(i)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            bytes_written = min((i + 1) * 20 * 1024 * 1024, local_image_size)
            print('    Progress: {:,} of {:,} bytes written: {:.2f}%'.format(bytes_written, local_image_size, bytes_written / local_image_size * 100), end='\r')
        print()
        print()
        print('The image has been burned to the USB drive!')

        retval = True

    except Exception as e:
        print('Error: {}'.format(e))

    return retval

################################################################################

def display_usb_boot_instructions(nta1000_system, offline):

    # All hardware versions have a hidden magic boot key that will bring up a
    # simple little boot screen that allows you to choose the boot device.
    # Only FWA8308 and FWA8408 boot keys seem to work consistently.
    # For FWA8308 (v5) and FWA8408 (v6,v7) systems, it's F7.

    print()
    print()
    print()
    print('The next step is to boot the system from the USB drive.')
    print()

    if offline:
        print('Power off the NTA1000 if not already powered off.')
        print('Once powered off, set the power switch on the back of the NTA1000')
        print('to the "off" position.  Leave it off for at least 10 seconds.')
        print('Unmount and remove the USB drive from this system,')
        print('and plug it into a USB port on the front of your NTA1000.')
        print('Set the power switch on the back of the NTA1000 to the "on" position.')
    else:
        print('Once the NTA1000 has powered off, set the power switch on the back')
        print('to the "off" position.  Leave it off for at least 10 seconds, then')
        print('turn it to the "on" position.')
    print()

    if nta1000_system in ['FWA8308', 'FWA8408']:
        print('As the NTA1000 boots, hold down the F7 key until you see the boot menu.')
        print('Select the menu option that corresponds to your USB drive.')
        print('The menu option text is determined by the manufacturer of your USB drive,')
        print('and therefore will differ from drive to drive.')
        print()
        print('IMPORTANT: Do NOT select any option starting with "UEFI".')
    elif nta1000_system == 'FWA8207':
        print('As the NTA1000 boots, hold down the DEL key until you enter the BIOS setup.')
        print('Once in the BIOS setup:')
        if nta1000_system == 'FWA8207':
            print('Navigate to: Boot -> Hard Disk Drives -> 1st Drive, Enter')
            print('Select the USB drive, Enter, Esc')
            print('Press F10 to save and exit')
        elif nta1000_system == 'FWA8206':
            print('Navigate to: Advanced BIOS Features -> Hard Disk Boot Priority, Enter')
            print('Set #1 to the USB drive')
            print('Press F10 to save and exit')        
    print()

    return

################################################################################

def download_progress(file, size, q):

    # We might not make it out of the gate if there's a problem starting the download.
    # Check for that first so no progress is shown.

    time.sleep(1)
    try:
        if q.get(block=False) == 'kill':
            return
    except queue.Empty:
        pass

    # Monitor the file we're downloading until we get a kill from the parent.
    while(True):
        if os.path.isfile(file):
            current_size = os.path.getsize(file)
            print('    Progress: {:,} of {:,} bytes downloaded: {:.2f}%'.format(current_size, size, current_size/size*100), end='\r')

        # See if we're done.
        try:
            if q.get(block=False) == 'kill':
                break
        except queue.Empty:
            pass

        time.sleep(1)

    print()

################################################################################

if __name__ == '__main__':
    main()
